@@ -4,19 +4,19 @@ And it really is simple.
 
 Setup::
 
-    >>> s = S3Bucket("mybucket", 
-    ...              access_key="ACESSS KEY",
-    ...              secret_key="SECRET KEY")
+    >>> s = S3Bucket(bucket,
+    ...              access_key=access_key,
+    ...              secret_key=secret_key)
     ... 
     >>> print s  # doctest: +ELLIPSIS
     <S3Bucket ... at 'https://s3.amazonaws.com/...'>
 
 or if you'd like to use virtual host S3::
 
-    >>> s = S3Bucket("mybucket",
-    ...              access_key="ACCESS KEY",
-    ...              secret_key="SECRET KEY",
-    ...              base_url="http://yo.se")
+    >>> s = S3Bucket(bucket,
+    ...              access_key=access_key,
+    ...              secret_key=secret_key,
+    ...              base_url=base_url)
     >>> print s  # doctest: +ELLIPSIS
     <S3Bucket ... at 'http...'>
 
@@ -72,10 +72,10 @@ If you just want to know about a key, ask and ye shall receive::
 
     >>> from pprint import pprint
     >>> s["This is a testfile."] = S3File("Hi!", metadata={"hairdo": "Secret"})
-    >>> pprint(s.info("test"))  # doctest: +ELLIPSIS
+    >>> pprint(s.info("This is a testfile."))  # doctest: +ELLIPSIS
     {'date': datetime.datetime(...),
      'headers': {'content-length': '3',
-                 'content-type': 'application/x-octet-stream',
+                 'content-type': 'application/octet-stream',
                  'date': '...',
                  'etag': '"..."',
                  'last-modified': '...',
@@ -84,7 +84,7 @@ If you just want to know about a key, ask and ye shall receive::
                  'x-amz-meta-hairdo': 'Secret',
                  'x-amz-request-id': '...'},
      'metadata': {'hairdo': 'Secret'},
-     'mimetype': 'application/x-octet-stream',
+     'mimetype': 'application/octet-stream',
      'modify': datetime.datetime(...),
      'size': 3}
 
@@ -181,12 +181,12 @@ def aws_urlquote(value):
     return urllib.quote(value, "/")
 
 def guess_mimetype(fn, default="application/octet-stream"):
-    """Guess a mimetype from filename *fn."""
+    """Guess a mimetype from filename *fn*."""
     if "." not in fn:
         return default
-    extension = fn.rsplit(".", 1)[1].lower()
-    if extension == "jpg": extension = "jpeg"
-    return mimetypes.guess_type("." + extension)[0] or default
+    bfn, ext = fn.lower().rsplit(".", 1)
+    if ext == "jpg": ext = "jpeg"
+    return mimetypes.guess_type(bfn + "." + ext)[0] or default
 
 def info_dict(headers):
    return {"size": int(headers["content-length"]),
@@ -208,14 +208,14 @@ def name(o):
 
     Classes:
     >>> name(Exception)
-    'Exception'
+    'exceptions.Exception'
     >>> class MyKlass(object): pass
     >>> name(MyKlass)
     'MyKlass'
 
     Instances:
     >>> name(Exception()), name(MyKlass())
-    ('Exception', 'MyKlass')
+    ('exceptions.Exception', 'MyKlass')
 
     Types:
     >>> name(str), name(object), name(int)
@@ -225,28 +225,34 @@ def name(o):
     >>> name("Hello"), name(True), name(None), name(Ellipsis)
     ('str', 'bool', 'NoneType', 'ellipsis')
     """
-    if hasattr(o, "__name__"): return o.__name__
-    for o in getattr(o, "__mro__", o.__class__.__mro__):
-        return name(o)
+    if hasattr(o, "__name__"):
+        rv = o.__name__
+        modname = getattr(o, "__module__", None)
+        # This work-around because Python does it itself,
+        # see typeobject.c, type_repr.
+        # Note that Python only checks for __builtin__.
+        if modname and modname[:2] + modname[-2:] != "____":
+            rv = o.__module__ + "." + rv
+    else:
+        for o in getattr(o, "__mro__", o.__class__.__mro__):
+            rv = name(o)
+            # If there is no name for the this baseclass, this ensures we check
+            # the next rather than say the object has no name (i.e., return
+            # None)
+            if rv is not None:
+                break
+    return rv
 
 class S3Error(Exception):
     def __init__(self, message, **kwds):
-        self.message = message
+        self.msg = message
         self.extra = kwds.copy()
 
     def __str__(self):
-        rv = self.message
+        rv = self.msg
         if self.extra:
-            key_it = iter(self.extra)
             rv += " ("
-            prep = ""
-            while len(rv) < 79:
-                try:
-                    key = key_it.next()
-                except StopIteration:
-                    break
-                rv += prep + key + "=" + repr(self.extra[key])
-                prep = ", "
+            rv += ", ".join("%s=%r" % i for i in self.extra.items())
             rv += ")"
         return rv
 
@@ -256,16 +262,17 @@ class S3Error(Exception):
         self = cls("HTTP error", code=e.code, url=e.filename)
         self.code = e.code
         self.fp = fp = e.fp
+        if not fp:
+            return self
         # The latter part of this clause is to avoid some weird bug in urllib2
         # and AWS which has it read as if chunked, and AWS gives empty reply.
-        if fp and dict(fp.info()).get("Content-Length", 0):
-            self.data = data = fp.read()
-            begin, end = data.find("<Message>"), data.find("</Message>")
-            if min(begin, end) >= 0:
-                self.full_message = msg = data[begin + 9:end]
-                self.message = msg[:50]
-                if self.message != msg:
-                    self.message += "..."
+        self.data = data = fp.read()
+        begin, end = data.find("<Message>"), data.find("</Message>")
+        if min(begin, end) >= 0:
+            self.full_message = msg = data[begin + 9:end]
+            self.msg = msg[:100]
+            if self.msg != msg:
+                self.msg += "..."
         return self
 
 class StreamHTTPHandler(urllib2.HTTPHandler):
@@ -339,13 +346,8 @@ class S3Bucket(object):
     def make_description(self, method, key=None, data=None,
                          headers={}, subresource=None, bucket=None):
         # The signature descriptor is detalied in the developer's PDF on p. 65.
-        # Calculate canonicalized resource.
-        res = "/"
-        if bucket or bucket is None:
-            res += aws_urlquote(bucket or self.name)
-        res += "/"
-        if key:
-            res += aws_urlquote(key)
+        res = self.canonicalized_resource(key, bucket=bucket)
+        # Append subresource, if any.
         if subresource:
             res += "?" + subresource
         # Make description. :/
@@ -353,14 +355,35 @@ class S3Bucket(object):
             headers.get("Content-Type", ""), headers.get("Date", ""))) + "\n" +\
             _amz_canonicalize(headers) + res
 
+    def canonicalized_resource(self, key, bucket=None):
+        res = "/"
+        if bucket or bucket is None:
+            res += aws_urlquote(bucket or self.name)
+        res += "/"
+        if key:
+            res += aws_urlquote(key)
+        return res
+
     def get_request_signature(self, method, key=None, data=None,
                               headers={}, subresource=None, bucket=None):
         return self.sign_description(self.make_description(method, key=key,
             data=data, headers=headers, subresource=subresource, bucket=bucket))
 
     def new_request(self, method, key=None, args=None, data=None, headers={}):
-        url = self.base_url + "/"
         headers = headers.copy()
+        if data and "Content-MD5" not in headers:
+            headers["Content-MD5"] = aws_md5(data)
+        if "Date" not in headers:
+            headers["Date"] = time.strftime(rfc822_fmt, time.gmtime())
+        if "Authorization" not in headers:
+            sign = self.get_request_signature(method, key=key, data=data,
+                                              headers=headers)
+            headers["Authorization"] = "AWS %s:%s" % (self.access_key, sign)
+        url = self.make_url(key, args)
+        return AnyMethodRequest(method, url, data=data, headers=headers)
+
+    def make_url(self, key, args, arg_sep=";"):
+        url = self.base_url + "/"
         if key:
             url += aws_urlquote(key)
         if args:
@@ -370,17 +393,9 @@ class S3Bucket(object):
                 args_items = args.items()
             else:
                 args_items = args
-            url += "?" + ";".join("=".join(map(urllib.quote_plus, item))
-                                  for item in args_items)
-        if data and "Content-MD5" not in headers:
-            headers["Content-MD5"] = aws_md5(data)
-        if "Date" not in headers:
-            headers["Date"] = time.strftime(rfc822_fmt, time.gmtime())
-        if "Authorization" not in headers:
-            sign = self.get_request_signature(method, key=key, data=data,
-                                              headers=headers)
-            headers["Authorization"] = "AWS %s:%s" % (self.access_key, sign)
-        return AnyMethodRequest(method, url, data=data, headers=headers)
+            url += "?" + arg_sep.join("=".join(map(urllib.quote_plus, item))
+                                      for item in args_items)
+        return url
 
     def open_request(self, request, errors=True):
             return self.opener.open(request)
@@ -438,11 +453,11 @@ class S3Bucket(object):
         else:
             if resp.code != 204:
                 raise KeyError(key)
-    
+
     def listdir(self, prefix=None, marker=None, limit=None, delimiter=None):
         """List contents of bucket.
 
-        *prefix*, if given, predicats `key.startswith(prefix)`.
+        *prefix*, if given, predicates `key.startswith(prefix)`.
         *marker*, if given, predicates `key > marker`, lexicographically.
         *limit*, if given, predicates `len(keys) <= limit`.
         """
@@ -473,6 +488,50 @@ class S3Bucket(object):
             if not data:
                 break
 
-if __name__ == "__main__":
-    import doctest
-    doctest.testmod()
+    def url_for(self, key, authenticated=False, expires=None,
+            method='GET', content_md5='', mimetype='',
+            headers={}):
+        """
+        Produces the URL for given S3 object key.
+
+        *key* specifies the S3 object path relative to the
+            base URL of the bucket.
+        *authenticated* asks for URL query string authentication
+            to be used; such URLs include a signature and expiration
+            time, and may be used by HTTP client apps to gain
+            temporary access to private S3 objects.
+        *expires* indicates when the produced URL ceases to function,
+            in seconds from Jan 1, 1970 UTC; if omitted, the URL
+            will expire in 5 minutes from now.
+
+        TODO: implement expires_in parameter that takes a
+            number of seconds from now or a positive timedelta object.
+
+        URL for a publicly accessible S3 object:
+        >>> S3Bucket('bottle').url_for('the dregs')
+        'https://s3.amazonaws.com/bottle/the%20dregs'
+
+        Query string authenitcated URL example (fake S3 credentials are shown):
+        >>> b = S3Bucket('johnsmith', access_key='0PN5J17HBGZHT7JJ3X82',
+        ...     secret_key='uV3F3YluFJax1cknvbcGwgjvx4QpvB+leU8dUj2o')
+        >>> b.url_for('foo.js', authenticated=True) # doctest: +ELLIPSIS
+        'https://s3.amazonaws.com/johnsmith/foo.js?AWSAccessKeyId=...&Expires=...&Signature=...'
+        """
+        args = tuple()
+        if authenticated:
+            expires = time.time() + 5 * 60 if expires is None else expires
+            expires = str(long(expires))
+            auth_descriptor = ''.join((
+                method, '\n',
+                content_md5, '\n',
+                mimetype, '\n',
+                expires, '\n',
+                _amz_canonicalize(headers), # No '\n' by design!
+                self.canonicalized_resource(key) # No '\n' by design!
+            ))
+            args = (
+                ('AWSAccessKeyId', self.access_key),
+                ('Expires', expires),
+                ('Signature', self.sign_description(auth_descriptor))
+            )
+        return self.make_url(key, args, '&')
